# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# 1. visit https://misskey-hub.net/ and see what it's all about if you want.
#    Choose a service you want to create an account on from here:
#    https://misskey-hub.net/en/instances.html
#
#    - For this plugin, I tested using https://misskey.sda1.net and created an
#      account.
#
# 2. Generate an API Key:
#    - Settings > API > Generate Key
#      - Name it whatever you want
#      - Assign it 'AT LEAST':
#          a. Compose or delete chat messages
#          b. Compose or delete notes
#
#
# This plugin also supports taking the URL (as identified above) directly
# as well.

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


class MisskeyVisibility:
    """The visibility of any note created."""

    # post will be public
    PUBLIC = "public"

    HOME = "home"

    FOLLOWERS = "followers"

    SPECIFIED = "specified"


# Define the types in a list for validation purposes
MISSKEY_VISIBILITIES = (
    MisskeyVisibility.PUBLIC,
    MisskeyVisibility.HOME,
    MisskeyVisibility.FOLLOWERS,
    MisskeyVisibility.SPECIFIED,
)


class NotifyMisskey(NotifyBase):
    """A wrapper for Misskey Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Misskey"

    # The services URL
    service_url = "https://misskey-hub.net/"

    # The default protocol
    protocol = "misskey"

    # The default secure protocol
    secure_protocol = "misskeys"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_misskey"

    # The title is not used
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 512

    # Define object templates
    templates = ("{schema}://{project_id}/{msghook}",)

    # Define object templates
    templates = (
        "{schema}://{token}@{host}",
        "{schema}://{token}@{host}:{port}",
    )

    # Define our template arguments
    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "token": {
                "name": _("Access Token"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "token": {
                "alias_of": "token",
            },
            "visibility": {
                "name": _("Visibility"),
                "type": "choice:string",
                "values": MISSKEY_VISIBILITIES,
                "default": MisskeyVisibility.PUBLIC,
            },
        },
    )

    def __init__(self, token=None, visibility=None, **kwargs):
        """Initialize Misskey Object."""
        super().__init__(**kwargs)

        self.token = validate_regex(token)
        if not self.token:
            msg = "An invalid Misskey Access Token was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if visibility:
            # Input is a string; attempt to get the lookup from our
            # sound mapping
            vis = (
                "invalid"
                if not isinstance(visibility, str)
                else visibility.lower().strip()
            )

            # This little bit of black magic allows us to match against
            # against multiple versions of the same string ... etc
            self.visibility = next(
                (v for v in MISSKEY_VISIBILITIES if v.startswith(vis)), None
            )

            if self.visibility not in MISSKEY_VISIBILITIES:
                msg = (
                    f"The Misskey visibility specified ({visibility}) is"
                    " invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.visibility = self.template_args["visibility"]["default"]

        # Prepare our URL
        self.schema = "https" if self.secure else "http"
        self.api_url = f"{self.schema}://{self.host}"

        if isinstance(self.port, int):
            self.api_url += f":{self.port}"

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.token,
            self.host,
            self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = {
            "visibility": self.visibility,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        host = self.host
        if isinstance(self.port, int):
            host += f":{self.port}"

        return "{schema}://{token}@{host}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            host=host,
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyMisskey.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Wrapper to _send since we can alert more then one channel."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Prepare our payload
        payload = {
            "i": self.token,
            "text": body,
            "visibility": self.visibility,
        }

        api_url = f"{self.api_url}/api/notes/create"
        self.logger.debug(
            "Misskey GET URL:"
            f" {api_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Misskey Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                api_url,
                headers=headers,
                data=dumps(payload),
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyMisskey.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Misskey notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Misskey notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Misskey notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyMisskey.unquote(results["qsd"]["token"])

        elif not results["password"] and results["user"]:
            results["token"] = NotifyMisskey.unquote(results["user"])

        # Capture visibility if specified
        if "visibility" in results["qsd"] and len(
            results["qsd"]["visibility"]
        ):
            results["visibility"] = NotifyMisskey.unquote(
                results["qsd"]["visibility"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://msg91.com/ if you don't already have one
#
# Get your (authkey) from the dashboard here:
#   - https://world.msg91.com/user/index.php#api
#
# Note: You will need to define a template for this to work
#
# Get details on the API used in this plugin here:
#   - https://docs.msg91.com/reference/send-sms
from json import dumps
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class MSG91PayloadField:
    """Identifies the fields available in the JSON Payload."""

    BODY = "body"
    MESSAGETYPE = "type"


# Add entries here that are reserved
RESERVED_KEYWORDS = ("mobiles",)


class NotifyMSG91(NotifyBase):
    """A wrapper for MSG91 Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "MSG91"

    # The services URL
    service_url = "https://msg91.com"

    # The default protocol
    secure_protocol = "msg91"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_msg91"

    # MSG91 uses the http protocol with JSON requests
    notify_url = "https://control.msg91.com/api/v5/flow/"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Our supported mappings and component keys
    component_key_re = re.compile(
        r"(?P<key>((?P<id>[a-z0-9_-])?|(?P<map>body|type)))", re.IGNORECASE
    )

    # Define object templates
    templates = ("{schema}://{template}@{authkey}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "template": {
                "name": _("Template ID"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9 _-]+$", "i"),
            },
            "authkey": {
                "name": _("Authentication Key"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "short_url": {
                "name": _("Short URL"),
                "type": "bool",
                "default": False,
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "template_mapping": {
            "name": _("Template Mapping"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        template,
        authkey,
        targets=None,
        short_url=None,
        template_mapping=None,
        **kwargs,
    ):
        """Initialize MSG91 Object."""
        super().__init__(**kwargs)

        # Authentication Key (associated with project)
        self.authkey = validate_regex(
            authkey, *self.template_tokens["authkey"]["regex"]
        )
        if not self.authkey:
            msg = (
                "An invalid MSG91 Authentication Key "
                f"({authkey}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Template ID
        self.template = validate_regex(
            template, *self.template_tokens["template"]["regex"]
        )
        if not self.template:
            msg = f"An invalid MSG91 Template ID ({template}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if short_url is None:
            self.short_url = self.template_args["short_url"]["default"]

        else:
            self.short_url = parse_bool(short_url)

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        self.template_mapping = {}
        if template_mapping:
            # Store our extra payload entries
            self.template_mapping.update(template_mapping)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform MSG91 Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no MSG91 targets to notify.")
            return False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "authkey": self.authkey,
        }

        # Base
        recipient_payload = {
            "mobiles": None,
            # Keyword Tokens
            MSG91PayloadField.BODY: body,
            MSG91PayloadField.MESSAGETYPE: notify_type,
        }

        # Prepare Recipient Payload Object
        for key, value in self.template_mapping.items():

            if key in RESERVED_KEYWORDS:
                self.logger.warning(
                    "Ignoring MSG91 custom payload entry %s", key
                )
                continue

            if key in recipient_payload:
                if not value:
                    # Do not store element in payload response
                    del recipient_payload[key]

                else:
                    # Re-map
                    recipient_payload[value] = recipient_payload[key]
                    del recipient_payload[key]

            else:
                # Append entry
                recipient_payload[key] = value

        # Prepare our recipients
        recipients = []
        for target in self.targets:
            recipient = recipient_payload.copy()
            recipient["mobiles"] = target
            recipients.append(recipient)

        # Prepare our payload
        payload = {
            "template_id": self.template,
            "short_url": 1 if self.short_url else 0,
            # target phone numbers are sent with a comma delimiter
            "recipients": recipients,
        }

        # Some Debug Logging
        self.logger.debug(
            "MSG91 POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"MSG91 Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyMSG91.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send MSG91 notification to {}: "
                    "{}{}error={}.".format(
                        ",".join(self.targets),
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return False

            else:
                self.logger.info(
                    "Sent MSG91 notification to {}.".format(
                        ",".join(self.targets)
                    )
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending MSG91:{} "
                "notification.".format(",".join(self.targets))
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.template, self.authkey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "short_url": str(self.short_url),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Payload body extras prefixed with a ':' sign
        # Append our payload extras into our parameters
        params.update({f":{k}": v for k, v in self.template_mapping.items()})

        return "{schema}://{template}@{authkey}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            template=self.pprint(self.template, privacy, safe=""),
            authkey=self.pprint(self.authkey, privacy, safe=""),
            targets="/".join(
                [NotifyMSG91.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyMSG91.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyMSG91.split_path(results["fullpath"])

        # The hostname is our authentication key
        results["authkey"] = NotifyMSG91.unquote(results["host"])

        # The template id is kept in the user field
        results["template"] = NotifyMSG91.unquote(results["user"])

        if "short_url" in results["qsd"] and len(results["qsd"]["short_url"]):
            results["short_url"] = parse_bool(results["qsd"]["short_url"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyMSG91.parse_phone_no(
                results["qsd"]["to"]
            )

        # store any additional payload extra's defined
        results["template_mapping"] = {
            NotifyMSG91.unquote(x): NotifyMSG91.unquote(y)
            for x, y in results["qsd:"].items()
        }

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
import re

import requests

from .. import exception
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase


class AppriseAPIMethod:
    """Defines the method to post data tot he remote server."""

    JSON = "json"
    FORM = "form"


APPRISE_API_METHODS = (
    AppriseAPIMethod.FORM,
    AppriseAPIMethod.JSON,
)


class NotifyAppriseAPI(NotifyBase):
    """A wrapper for Apprise (Persistent) API Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Apprise API"

    # The services URL
    service_url = "https://github.com/caronc/apprise-api"

    # The default protocol
    protocol = "apprise"

    # The default secure protocol
    secure_protocol = "apprises"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_apprise_api"

    # Support attachments
    attachment_support = True

    # Depending on the number of transactions/notifications taking place, this
    # could take a while. 30 seconds should be enough to perform the task
    socket_read_timeout = 30.0

    # Disable throttle rate for Apprise API requests since they are normally
    # local anyway
    request_rate_per_sec = 0.0

    # Define object templates
    templates = (
        "{schema}://{host}/{token}",
        "{schema}://{host}:{port}/{token}",
        "{schema}://{user}@{host}/{token}",
        "{schema}://{user}@{host}:{port}/{token}",
        "{schema}://{user}:{password}@{host}/{token}",
        "{schema}://{user}:{password}@{host}:{port}/{token}",
    )

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "token": {
                "name": _("Token"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[A-Z0-9_-]{1,128}$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "tags": {
                "name": _("Tags"),
                "type": "string",
            },
            "method": {
                "name": _("Query Method"),
                "type": "choice:string",
                "values": APPRISE_API_METHODS,
                "default": APPRISE_API_METHODS[0],
            },
            "to": {
                "alias_of": "token",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(
        self, token=None, tags=None, method=None, headers=None, **kwargs
    ):
        """Initialize Apprise API Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"The Apprise API token specified ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.method = (
            self.template_args["method"]["default"]
            if not isinstance(method, str)
            else method.lower()
        )

        if self.method not in APPRISE_API_METHODS:
            msg = f"The method specified ({method}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Build list of tags
        self.__tags = parse_list(tags)

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "method": self.method,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        if self.__tags:
            params["tags"] = ",".join(list(self.__tags))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyAppriseAPI.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyAppriseAPI.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80

        fullpath = self.fullpath.strip("/")
        return (
            "{schema}://{auth}{hostname}{port}{fullpath}{token}"
            "/?{params}".format(
                schema=self.secure_protocol if self.secure else self.protocol,
                auth=auth,
                # never encode hostname since we're expecting it to be a
                # valid one
                hostname=self.host,
                port=(
                    ""
                    if self.port is None or self.port == default_port
                    else f":{self.port}"
                ),
                fullpath=(
                    "/{}/".format(NotifyAppriseAPI.quote(fullpath, safe="/"))
                    if fullpath
                    else "/"
                ),
                token=self.pprint(self.token, privacy, safe=""),
                params=NotifyAppriseAPI.urlencode(params),
            )
        )

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Apprise API Notification."""

        # Prepare HTTP Headers
        headers = {
            "User-Agent": self.app_id,
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        attachments = []
        files = []
        if attach and self.attachment_support:
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Apprise API attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    # Our Attachment filename
                    filename = (
                        attachment.name
                        if attachment.name
                        else f"file{no:03}.dat"
                    )

                    if self.method == AppriseAPIMethod.JSON:
                        # Output must be in a DataURL format (that's what
                        # PushSafer calls it):
                        attachments.append({
                            "filename": filename,
                            "base64": attachment.base64(),
                            "mimetype": attachment.mimetype,
                        })

                    else:  # AppriseAPIMethod.FORM
                        files.append((
                            f"file{no:02d}",
                            (
                                filename,
                                # file handle is safely closed in `finally`;
                                # inline open is intentional
                                open(attachment.path, "rb"),  # noqa: SIM115
                                attachment.mimetype,
                            ),
                        ))

                except (TypeError, OSError, exception.AppriseException):
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access AppriseAPI attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending AppriseAPI attachment"
                    f" {attachment.url(privacy=True)}"
                )

        # prepare Apprise API Object
        payload = {
            # Apprise API Payload
            "title": title,
            "body": body,
            "type": notify_type.value,
            "format": self.notify_format.value,
        }

        if self.method == AppriseAPIMethod.JSON:
            headers["Content-Type"] = "application/json"

            if attachments:
                payload["attachments"] = attachments

            payload = dumps(payload)

        if self.__tags:
            payload["tag"] = self.__tags

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        fullpath = self.fullpath.strip("/")
        url += "{}".format("/" + fullpath) if fullpath else ""
        url += f"/notify/{self.token}"

        # Some entries can not be over-ridden
        headers.update({
            # Our response to be in JSON format always
            "Accept": "application/json",
            # Pass our Source UUID4 Identifier
            "X-Apprise-ID": self.asset._uid,
            # Pass our current recursion count to our upstream server
            "X-Apprise-Recursion-Count": str(self.asset._recursion + 1),
        })

        self.logger.debug(
            "Apprise API POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Apprise API Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=payload,
                headers=headers,
                auth=auth,
                files=files if files else None,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyAppriseAPI.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Apprise API notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info(
                    "Sent Apprise API notification; method=%s.", self.method
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Apprise API "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading one of the "
                "attached files."
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            for file in files:
                # Ensure all files are closed
                file[1][1].close()

        return True

    @staticmethod
    def parse_native_url(url):
        """
        Support http://hostname/notify/token and
                http://hostname/path/notify/token
        """

        result = re.match(
            r"^http(?P<secure>s?)://(?P<hostname>[A-Z0-9._-]+)"
            r"(:(?P<port>[0-9]+))?"
            r"(?P<path>/[^?]+?)?/notify/(?P<token>[A-Z0-9_-]{1,32})/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyAppriseAPI.parse_url(
                "{schema}://{hostname}{port}{path}/{token}/{params}".format(
                    schema=(
                        NotifyAppriseAPI.secure_protocol
                        if result.group("secure")
                        else NotifyAppriseAPI.protocol
                    ),
                    hostname=result.group("hostname"),
                    port=(
                        ""
                        if not result.group("port")
                        else ":{}".format(result.group("port"))
                    ),
                    path=(
                        ""
                        if not result.group("path")
                        else result.group("path")
                    ),
                    token=result.group("token"),
                    params=(
                        ""
                        if not result.group("params")
                        else "?{}".format(result.group("params"))
                    ),
                )
            )

        return None

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyAppriseAPI.unquote(x): NotifyAppriseAPI.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Support the passing of tags in the URL
        if "tags" in results["qsd"] and len(results["qsd"]["tags"]):
            results["tags"] = NotifyAppriseAPI.parse_list(
                results["qsd"]["tags"]
            )

        # Support the 'to' & 'token' variable so that we can support rooms
        # this way too.
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyAppriseAPI.unquote(
                results["qsd"]["token"]
            )

        elif "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["token"] = NotifyAppriseAPI.unquote(results["qsd"]["to"])

        else:
            # Start with a list of path entries to work with
            entries = NotifyAppriseAPI.split_path(results["fullpath"])
            if entries:
                # use our last entry found
                results["token"] = entries[-1]

                # pop our last entry off
                entries = entries[:-1]

                # re-assemble our full path
                results["fullpath"] = "/".join(entries)

        # Set method if specified
        if "method" in results["qsd"] and len(results["qsd"]["method"]):
            results["method"] = NotifyAppriseAPI.unquote(
                results["qsd"]["method"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timedelta, timezone
import hashlib
import os

from ..apprise_attachment import AppriseAttachment
from ..asset import AppriseAsset
from ..exception import ApprisePluginException
from ..logger import logger

try:
    import pgpy

    # Pretty Good Privacy (PGP) Support enabled
    PGP_SUPPORT = True

except ImportError:
    # Pretty Good Privacy (PGP) Support disabled
    PGP_SUPPORT = False


class ApprisePGPException(ApprisePluginException):
    """Thrown when there is an error with the Pretty Good Privacy
    Controller."""

    def __init__(self, message, error_code=602):
        super().__init__(message, error_code=error_code)


class ApprisePGPController:
    """Pretty Good Privacy Controller Tool for the Apprise Library."""

    # There is no reason a PGP Public Key should exceed 8K in size
    # If it is more than this, then it is not accepted
    max_pgp_public_key_size = 8000

    def __init__(
        self, path, pub_keyfile=None, email=None, asset=None, **kwargs
    ):
        """Path should be the directory keys can be written and read from such
        as <notifyobject>.store.path.

        Optionally additionally specify a keyfile to explicitly open
        """

        # PGP hash
        self.__key_lookup = {}

        # Directory we can work with
        self.path = path

        # Our email
        self.email = email

        # Prepare our Asset Object
        self.asset = (
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()
        )

        if pub_keyfile:
            # Create ourselves an Attachment to work with; this grants us the
            # ability to pull this key from a remote site or anything else
            # supported by the Attachment object
            self._pub_keyfile = AppriseAttachment(asset=self.asset)

            # Add our definition to our pgp_key reference
            self._pub_keyfile.add(pub_keyfile)

            # Enforce maximum file size
            self._pub_keyfile[0].max_file_size = self.max_pgp_public_key_size

        else:
            self._pub_keyfile = None

    def keygen(self, email=None, name=None, force=False):
        """Generates a set of keys based on email configured."""

        try:
            # Create a new RSA key pair with 2048-bit strength
            key = pgpy.PGPKey.new(
                pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048
            )

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; keygen disabled")
            return False

        if self._pub_keyfile is not None or not self.path:
            logger.trace(
                "PGP keygen disabled, reason=%s",
                (
                    "keyfile-defined"
                    if self._pub_keyfile is not None
                    else "no-write-path"
                ),
            )
            return False

        if not name:
            name = self.asset.app_id

        if not email:
            email = self.email

        # Prepare our UID
        uid = pgpy.PGPUID.new(name, email=email)

        # Filenames
        file_prefix = email.split("@")[0].lower()

        pub_path = os.path.join(self.path, f"{file_prefix}-pub.asc")
        prv_path = os.path.join(self.path, f"{file_prefix}-prv.asc")

        if os.path.isfile(pub_path) and not force:
            logger.debug(
                "PGP generation skipped; Public Key already exists: %s",
                pub_path,
            )
            return True

        # Persistent Storage Key
        lookup_key = hashlib.sha1(
            os.path.abspath(pub_path).encode("utf-8")
        ).hexdigest()
        if lookup_key in self.__key_lookup:
            # Ensure our key no longer exists
            del self.__key_lookup[lookup_key]

        # Add the user ID to the key
        key.add_uid(
            uid,
            usage={
                pgpy.constants.KeyFlags.Sign,
                pgpy.constants.KeyFlags.EncryptCommunications,
            },
            hashes=[pgpy.constants.HashAlgorithm.SHA256],
            ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
            compression=[pgpy.constants.CompressionAlgorithm.ZLIB],
        )

        try:
            # Write our keys to disk
            with open(pub_path, "w") as f:
                f.write(str(key.pubkey))

        except OSError as e:
            logger.warning("Error writing PGP file %s", pub_path)
            logger.debug(f"I/O Exception: {e}")

            # Cleanup
            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

        try:
            with open(prv_path, "w") as f:
                f.write(str(key))

        except OSError as e:
            logger.warning("Error writing PGP file %s", prv_path)
            logger.debug(f"I/O Exception: {e}")

            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

            try:
                os.unlink(prv_path)
                logger.trace("Removed %s", prv_path)

            except OSError:
                pass

            return False

        logger.info(
            "Wrote PGP Keys for %s/%s",
            os.path.dirname(pub_path),
            os.path.basename(pub_path),
        )
        return True

    def public_keyfile(self, *emails):
        """Returns the first match of a useable public key based emails
        provided."""

        if not PGP_SUPPORT:
            msg = "PGP Support unavailable; install PGPy library"
            logger.warning(msg)
            raise ApprisePGPException(msg)

        if self._pub_keyfile is not None:
            # If our code reaches here, then we fetch our public key
            pgp_key = self._pub_keyfile[0]
            if not pgp_key:
                # We could not access the attachment
                logger.error(
                    "Could not access PGP Public Key"
                    f" {pgp_key.url(privacy=True)}."
                )
                return False

            return pgp_key.path

        elif not self.path:
            # No path
            return None

        fnames = [
            "pgp-public.asc",
            "pgp-pub.asc",
            "public.asc",
            "pub.asc",
        ]

        if self.email:
            # Include our email in the list
            emails = [self.email, *emails]

        for email in emails:
            _entry = email.split("@")[0].lower()
            fnames.insert(0, f"{_entry}-pub.asc")

            # Lowercase email (Highest Priority)
            _entry = email.lower()
            fnames.insert(0, f"{_entry}-pub.asc")

        return next(
            (
                os.path.join(self.path, fname)
                for fname in fnames
                if os.path.isfile(os.path.join(self.path, fname))
            ),
            None,
        )

    def public_key(self, *emails, autogen=None):
        """Opens a spcified pgp public file and returns the key from it which
        is used to encrypt the message."""
        path = self.public_keyfile(*emails)
        if not path:
            if (
                autogen if autogen is not None else self.asset.pgp_autogen
            ) and self.keygen(*emails):
                path = self.public_keyfile(*emails)
                if path:
                    # We should get a hit now
                    return self.public_key(*emails)

            logger.warning("No PGP Public Key could be loaded")
            return None

        # Persistent Storage Key
        key = hashlib.sha1(os.path.abspath(path).encode("utf-8")).hexdigest()
        if key in self.__key_lookup:
            # Take an early exit
            return self.__key_lookup[key]["public_key"]

        try:
            with open(path) as key_file:
                public_key, _ = pgpy.PGPKey.from_blob(key_file.read())

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; skipping PGP support: %s", path)
            return None

        except FileNotFoundError:
            # Generate keys
            logger.debug("PGP Public Key file not found: %s", path)
            return None

        except OSError as e:
            logger.warning("Error accessing PGP Public Key file %s", path)
            logger.debug(f"I/O Exception: {e}")
            return None

        self.__key_lookup[key] = {
            "public_key": public_key,
            "expires": datetime.now(timezone.utc) + timedelta(seconds=86400),
        }
        return public_key

    # Encrypt message using the recipient's public key
    def encrypt(self, message, *emails):
        """If provided a path to a pgp-key, content is encrypted."""

        # Acquire our key
        public_key = self.public_key(*emails)
        if not public_key:
            # Encryption not possible
            return False

        try:
            message_object = pgpy.PGPMessage.new(message)
            encrypted_message = public_key.encrypt(message_object)
            return str(encrypted_message)

        except pgpy.errors.PGPError:
            # Encryption not Possible
            logger.debug("PGP Public Key Corruption; encryption not possible")

        except NameError:
            # PGPy not installed
            logger.debug("PGPy not installed; Skipping PGP encryption")

        return None

    def prune(self):
        """Prunes old entries from the public_key index."""
        self.__key_lookup = {
            key: value
            for key, value in self.__key_lookup.items()
            if value["expires"] > datetime.now(timezone.utc)
        }

    @property
    def pub_keyfile(self):
        """Returns the Public Keyfile Path if set otherwise it returns None
        This property returns False if a keyfile was provided, but was
        invalid."""
        return (
            None
            if self._pub_keyfile is None
            else (
                False
                if not self._pub_keyfile[0]
                else self._pub_keyfile[0].path
            )
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a Kavenegar account from their website
# at https://kavenegar.com/
#
# After you've established your account you can get your API Key from your
# account profile: https://panel.kavenegar.com/client/setting/account
#
# This provider does not accept +1 (for example) as a country code. You need
# to specify 001 instead.
#
from json import loads

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
# Based on https://kavenegar.com/rest.html
KAVENEGAR_HTTP_ERROR_MAP = {
    200: "The request was approved",
    400: "Parameters are incomplete",
    401: "Account has been disabled",
    402: "The operation failed",
    403: "The API Key is invalid",
    404: "The method is unknown",
    405: "The GET/POST request is wrong",
    406: "Invalid mandatory parameters sent",
    407: "You canot access the information you want",
    409: "The server is unable to response",
    411: "The recipient is invalid",
    412: "The sender is invalid",
    413: "Message empty or message length exceeded",
    414: "The number of recipients is more than 200",
    415: "The start index is larger then the total",
    416: "The source IP of the service does not match the settings",
    417: (
        "The submission date is incorrect, "
        "either expired or not in the correct format"
    ),
    418: "Your account credit is insufficient",
    422: "Data cannot be processed due to invalid characters",
    501: "SMS can only be sent to the account holder number",
}


class NotifyKavenegar(NotifyBase):
    """A wrapper for Kavenegar Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Kavenegar"

    # The services URL
    service_url = "https://kavenegar.com/"

    # All notification requests are secure
    secure_protocol = "kavenegar"

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.20

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_kavenegar"

    # Kavenegar single notification URL
    notify_url = "http://api.kavenegar.com/v1/{apikey}/sms/send.json"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{apikey}/{targets}",
        "{schema}://{source}@{apikey}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "source": {
                "name": _("Source Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "source",
            },
        },
    )

    def __init__(self, apikey, source=None, targets=None, **kwargs):
        """Initialize Kavenegar Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Kavenegar API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.source = None
        if source is not None:
            result = is_phone_no(source)
            if not result:
                msg = f"The Kavenegar source specified ({source}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)

            # Store our source
            self.source = result["full"]

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Sends SMS Message."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no Kavenegar targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Accept": "application/json",
        }

        # Our URL
        url = self.notify_url.format(apikey=self.apikey)

        # use the list directly
        targets = list(self.targets)

        while len(targets):
            # Get our target(s) to notify
            target = targets.pop(0)

            # Prepare our payload
            payload = {
                "receptor": target,
                "message": body,
            }

            if self.source:
                # Only set source if specified
                payload["sender"] = self.source

            # Some Debug Logging
            self.logger.debug(
                "Kavenegar POST URL:"
                f" {url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Kavenegar Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    url,
                    params=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code not in (
                    requests.codes.created,
                    requests.codes.ok,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code, KAVENEGAR_HTTP_ERROR_MAP
                    )

                    try:
                        # Update our status response if we can
                        json_response = loads(r.content)
                        status_str = json_response.get("message", status_str)

                    except (AttributeError, TypeError, ValueError):
                        # ValueError = r.content is Unparsable
                        # TypeError = r.content is None
                        # AttributeError = r is None

                        # We could not parse JSON response.
                        # We will just use the status we already have.
                        pass

                    self.logger.warning(
                        "Failed to send Kavenegar SMS notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                # If we reach here; the message was sent
                self.logger.info(
                    f"Sent Kavenegar SMS notification to {target}."
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Kavenegar:{} ".format(
                        ", ".join(self.targets)
                    )
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.source, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{source}{apikey}/{targets}?{params}".format(
            schema=self.secure_protocol,
            source="" if not self.source else f"{self.source}@",
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join(
                [NotifyKavenegar.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyKavenegar.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store the source if specified
        if results.get("user", None):
            results["source"] = results["user"]

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyKavenegar.split_path(results["fullpath"])

        # The hostname is our authentication key
        results["apikey"] = NotifyKavenegar.unquote(results["host"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyKavenegar.parse_phone_no(
                results["qsd"]["to"]
            )

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyKavenegar.unquote(results["qsd"]["from"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Used to break path apart into list of targets
TARGET_LIST_DELIM = re.compile(r"[ \t\r\n,\\/]+")


# Priorities
class ParsePlatformDevice:
    # All Devices
    ALL = "all"

    # Apple IOS (APNS)
    IOS = "ios"

    # Android/Firebase (FCM)
    ANDROID = "android"


PARSE_PLATFORM_DEVICES = (
    ParsePlatformDevice.ALL,
    ParsePlatformDevice.IOS,
    ParsePlatformDevice.ANDROID,
)


class NotifyParsePlatform(NotifyBase):
    """A wrapper for Parse Platform Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Parse Platform"

    # The services URL
    service_url = " https://parseplatform.org/"

    # insecure notifications (using http)
    protocol = "parsep"

    # Secure notifications (using https)
    secure_protocol = "parseps"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_parseplatform"

    # Define object templates
    templates = (
        "{schema}://{app_id}:{master_key}@{host}",
        "{schema}://{app_id}:{master_key}@{host}:{port}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "app_id": {
                "name": _("App ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "master_key": {
                "name": _("Master Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "device": {
                "name": _("Device"),
                "type": "choice:string",
                "values": PARSE_PLATFORM_DEVICES,
                "default": ParsePlatformDevice.ALL,
            },
            "app_id": {
                "alias_of": "app_id",
            },
            "master_key": {
                "alias_of": "master_key",
            },
        },
    )

    def __init__(self, app_id, master_key, device=None, **kwargs):
        """Initialize Parse Platform Object."""
        super().__init__(**kwargs)

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = "/"

        # Application ID
        self.application_id = validate_regex(app_id)
        if not self.application_id:
            msg = (
                "An invalid Parse Platform Application ID "
                f"({app_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Master Key
        self.master_key = validate_regex(master_key)
        if not self.master_key:
            msg = (
                "An invalid Parse Platform Master Key "
                f"({master_key}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Initialize Devices Array
        self.devices = []

        if device:
            self.device = device.lower()
            if device not in PARSE_PLATFORM_DEVICES:
                msg = (
                    "An invalid Parse Platform device "
                    f"({device}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.device = self.template_args["device"]["default"]

        if self.device == ParsePlatformDevice.ALL:
            self.devices = [
                d
                for d in PARSE_PLATFORM_DEVICES
                if d != ParsePlatformDevice.ALL
            ]
        else:
            # Store our device
            self.devices.append(device)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Parse Platform Notification."""

        # Prepare our headers:
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Parse-Application-Id": self.application_id,
            "X-Parse-Master-Key": self.master_key,
        }

        # prepare our payload
        payload = {
            "where": {
                "deviceType": {
                    "$in": self.devices,
                }
            },
            "data": {
                "title": title,
                "alert": body,
            },
        }

        # Set our schema
        schema = "https" if self.secure else "http"

        # Our Notification URL
        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += self.fullpath.rstrip("/") + "/parse/push/"

        self.logger.debug(
            "Parse Platform POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Parse Platform Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyParsePlatform.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Parse Platform notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Parse Platform notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occured sending Parse Platform "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.application_id,
            self.master_key,
            self.host,
            self.port,
            self.fullpath,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any arguments set
        params = {
            "device": self.device,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        default_port = 443 if self.secure else 80

        return (
            "{schema}://{app_id}:{master_key}@"
            "{hostname}{port}{fullpath}/?{params}".format(
                schema=self.secure_protocol if self.secure else self.protocol,
                app_id=self.pprint(self.application_id, privacy, safe=""),
                master_key=self.pprint(self.master_key, privacy, safe=""),
                hostname=NotifyParsePlatform.quote(self.host, safe=""),
                port=(
                    ""
                    if self.port is None or self.port == default_port
                    else f":{self.port}"
                ),
                fullpath=NotifyParsePlatform.quote(self.fullpath, safe="/"),
                params=NotifyParsePlatform.urlencode(params),
            )
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        substantiate this object."""
        results = NotifyBase.parse_url(url)

        if not results:
            # We're done early as we couldn't load the results
            return results

        # App ID is retrieved from the user
        results["app_id"] = NotifyParsePlatform.unquote(results["user"])

        # Master Key is retrieved from the password
        results["master_key"] = NotifyParsePlatform.unquote(
            results["password"]
        )

        # Device support override
        if "device" in results["qsd"] and len(results["qsd"]["device"]):
            results["device"] = results["qsd"]["device"]

        # Allow app_id attribute over-ride
        if "app_id" in results["qsd"] and len(results["qsd"]["app_id"]):
            results["app_id"] = results["qsd"]["app_id"]

        # Allow master_key attribute over-ride
        if "master_key" in results["qsd"] and len(
            results["qsd"]["master_key"]
        ):
            results["master_key"] = results["qsd"]["master_key"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://gateway.threema.ch/en/ if you don't already have
# one
#
# Read more about Threema Gateway API here:
#   - https://gateway.threema.ch/en/developer/api

from itertools import chain

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_email, is_phone_no, parse_list, validate_regex
from .base import NotifyBase


class ThreemaRecipientTypes:
    """The supported recipient specifiers."""

    THREEMA_ID = "to"
    PHONE = "phone"
    EMAIL = "email"


class NotifyThreema(NotifyBase):
    """A wrapper for Threema Gateway Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Threema Gateway"

    # The services URL
    service_url = "https://gateway.threema.ch/"

    # The default protocol
    secure_protocol = "threema"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_threema"

    # Threema Gateway uses the http protocol with JSON requests
    notify_url = "https://msgapi.threema.ch/send_simple"

    # The maximum length of the body
    body_maxlen = 3500

    # No title support
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{gateway_id}@{secret}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "gateway_id": {
                "name": _("Gateway ID"),
                "type": "string",
                "private": True,
                "required": True,
                "map_to": "user",
            },
            "secret": {
                "name": _("API Secret"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "target_threema_id": {
                "name": _("Target Threema ID"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "gateway_id",
            },
            "gwid": {
                "alias_of": "gateway_id",
            },
            "secret": {
                "alias_of": "secret",
            },
        },
    )

    def __init__(self, secret=None, targets=None, **kwargs):
        """Initialize Threema Gateway Object."""
        super().__init__(**kwargs)

        # Validate our params here.

        if not self.user:
            msg = "Threema Gateway ID must be specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Verify our Gateway ID
        if len(self.user) != 8:
            msg = "Threema Gateway ID must be 8 characters in length"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Verify our secret
        self.secret = validate_regex(secret)
        if not self.secret:
            msg = f"An invalid Threema API Secret ({secret}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Parse our targets
        self.targets = []

        # Used for URL generation afterwards only
        self.invalid_targets = []

        for target in parse_list(targets, allow_whitespace=False):
            if len(target) == 8:
                # Store our user
                self.targets.append((ThreemaRecipientTypes.THREEMA_ID, target))
                continue

            # Check if an email was defined
            result = is_email(target)
            if result:
                # Store our user
                self.targets.append(
                    (ThreemaRecipientTypes.EMAIL, result["full_email"])
                )
                continue

            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if result:
                # store valid phone number
                self.targets.append(
                    (ThreemaRecipientTypes.PHONE, result["full"])
                )
                continue

            self.logger.warning(
                f"Dropped invalid user/email/phone ({target}) specified",
            )
            self.invalid_targets.append(target)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Threema Gateway Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning(
                "There were no Threema Gateway targets to notify"
            )
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
            "Accept": "*/*",
        }

        # Prepare our payload
        _payload = {
            "secret": self.secret,
            "from": self.user,
            "text": body.encode("utf-8"),
        }

        # Create a copy of the targets list
        targets = list(self.targets)

        while len(targets):
            # Get our target to notify
            key, target = targets.pop(0)

            # Prepare a payload object
            payload = _payload.copy()

            # Set Target
            payload[key] = target

            # Some Debug Logging
            self.logger.debug(
                "Threema Gateway GET URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Threema Gateway Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    self.notify_url,
                    params=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyThreema.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Threema Gateway notification to {}: "
                        "{}{}error={}".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                # We wee successful
                self.logger.info(
                    f"Sent Threema Gateway notification to {target}"
                )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Threema"
                    f" Gateway:{target} notification"
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user, self.secret)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        schemaStr = "{schema}://{gatewayid}@{secret}/{targets}?{params}"
        return schemaStr.format(
            schema=self.secure_protocol,
            gatewayid=NotifyThreema.quote(self.user),
            secret=self.pprint(
                self.secret, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            targets="/".join(
                chain(
                    [
                        NotifyThreema.quote(x[1], safe="@+")
                        for x in self.targets
                    ],
                    [
                        NotifyThreema.quote(x, safe="@+")
                        for x in self.invalid_targets
                    ],
                )
            ),
            params=NotifyThreema.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        results["targets"] = []

        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            results["secret"] = NotifyThreema.unquote(results["qsd"]["secret"])

        else:
            results["secret"] = NotifyThreema.unquote(results["host"])

        results["targets"] += NotifyThreema.split_path(results["fullpath"])

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["user"] = NotifyThreema.unquote(results["qsd"]["from"])

        elif "gwid" in results["qsd"] and len(results["qsd"]["gwid"]):
            results["user"] = NotifyThreema.unquote(results["qsd"]["gwid"])

        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyThreema.parse_list(
                results["qsd"]["to"], allow_whitespace=False
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, tzinfo
from os.path import abspath, dirname, isfile, join
import re
from typing import Any, Optional, Union
from uuid import uuid4

from .common import (
    NotifyFormat,
    NotifyImageSize,
    NotifyType,
    PersistentStoreMode,
)
from .manager_plugins import NotificationManager
from .utils.time import zoneinfo

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


class AppriseAsset:
    """Provides a supplimentary class that can be used to provide extra
    information and details that can be used by Apprise such as providing an
    alternate location to where images/icons can be found and the URL masks.

    Any variable that starts with an underscore (_) can only be initialized by
    this class manually and will/can not be parsed from a configuration file.
    """

    # Application Identifier
    app_id = "Apprise"

    # Application Description
    app_desc = "Apprise Notifications"

    # Provider URL
    app_url = "https://github.com/caronc/apprise"

    # A Simple Mapping of Colors; For every NOTIFY_TYPE identified,
    # there should be a mapping to it's color here:
    html_notify_map = {
        NotifyType.INFO: "#3AA3E3",
        NotifyType.SUCCESS: "#3AA337",
        NotifyType.FAILURE: "#A32037",
        NotifyType.WARNING: "#CACF29",
    }

    # The default color to return if a mapping isn't found in our table above
    default_html_color = "#888888"

    # Ascii Notification
    ascii_notify_map = {
        NotifyType.INFO: "[i]",
        NotifyType.SUCCESS: "[+]",
        NotifyType.FAILURE: "[!]",
        NotifyType.WARNING: "[~]",
    }

    # The default ascii to return if a mapping isn't found in our table above
    default_ascii_chars = "[?]"

    # The default image extension to use
    default_extension = ".png"

    # The default image size if one isn't specified
    default_image_size = NotifyImageSize.XY_256

    # The default theme
    theme = "default"

    # Image URL Mask
    image_url_mask = (
        "https://github.com/caronc/apprise/raw/master/apprise/assets/"
        "themes/{THEME}/apprise-{TYPE}-{XY}{EXTENSION}"
    )

    # Application Logo
    image_url_logo = (
        "https://github.com/caronc/apprise/raw/master/apprise/assets/"
        "themes/{THEME}/apprise-logo.png"
    )

    # Image Path Mask
    image_path_mask = abspath(
        join(
            dirname(__file__),
            "assets",
            "themes",
            "{THEME}",
            "apprise-{TYPE}-{XY}{EXTENSION}",
        )
    )

    # This value can also be set on calls to Apprise.notify(). This allows
    # you to let Apprise upfront the type of data being passed in.  This
    # must be of type NotifyFormat. Possible values could be:
    # - NotifyFormat.TEXT
    # - NotifyFormat.MARKDOWN
    # - NotifyFormat.HTML
    # - None
    #
    # If no format is specified (hence None), then no special pre-formatting
    # actions will take place during a notification. This has been and always
    # will be the default.
    body_format = None

    # Always attempt to send notifications asynchronous (as the same time
    # if possible)
    # This is a Python 3 supported option only. If set to False, then
    # notifications are sent sequentially (one after another)
    async_mode = True

    # Support :smile:, and other alike keywords swapping them for their
    # unicode value. A value of None leaves the interpretation up to the
    # end user to control (allowing them to specify emojis=yes on the
    # URL)
    interpret_emojis = None

    # Whether or not to interpret escapes found within the input text prior
    # to passing it upstream. Such as converting \t to an actual tab and \n
    # to a new line.
    interpret_escapes = False

    # Defines the encoding of the content passed into Apprise
    encoding = "utf-8"

    # Automatically generate our Pretty Good Privacy (PGP) keys if one isn't
    # present and our environment configuration allows for it.
    # For example, a case where the environment wouldn't allow for it would be
    # if Persistent Storage was set to `memory`
    pgp_autogen = True

    # Automatically generate our Privacy Enhanced Mail (PEM) keys if one isn't
    # present and our environment configuration allows for it.
    # For example, a case where the environment wouldn't allow for it would be
    # if Persistent Storage was set to `memory`
    pem_autogen = True

    # For more detail see CWE-312 @
    #    https://cwe.mitre.org/data/definitions/312.html
    #
    # By enabling this, the logging output has additional overhead applied to
    # it preventing secure password and secret information from being
    # displayed in the logging. Since there is overhead involved in performing
    # this cleanup; system owners who run in a very isolated environment may
    # choose to disable this for a slight performance bump. It is recommended
    # that you leave this option as is otherwise.
    secure_logging = True

    # Optionally specify one or more path to attempt to scan for Python modules
    # By default, no paths are scanned.
    __plugin_paths = []

    # Optionally set the location of the persistent storage
    # By default there is no path and thus persistent storage is not used
    __storage_path = None

    # Optionally define the default salt to apply to all persistent storage
    # namespace generation (unless over-ridden)
    __storage_salt = b""

    # Optionally define the namespace length of the directories created by
    # the storage. If this is set to zero, then the length is pre-determined
    # by the generator (sha1, md5, sha256, etc)
    __storage_idlen = 8

    # Set storage to auto
    __storage_mode = PersistentStoreMode.AUTO

    # All internal/system flags are prefixed with an underscore (_)
    # These can only be initialized using Python libraries and are not picked
    # up from (yaml) configuration files (if set)

    # An internal counter that is used by AppriseAPI
    # (https://github.com/caronc/apprise-api). The idea is to allow one
    # instance of AppriseAPI to call another, but to track how many times
    # this occurs. It's intent is to prevent a loop where an AppriseAPI
    # Server calls itself (or loops indefinitely)
    _recursion = 0

    # A unique identifer we can use to associate our calling source
    _uid = str(uuid4())

    # Default timezone to use (pass in timezone value)
    # A list of timezones can be found here:
    # https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    # You can specify things such as 'America/Montreal'
    # If no timezone is specified, then the one detected on the system
    # is uzed
    _tzinfo = None

    def __init__(
        self,
        plugin_paths: Optional[list[str]] = None,
        storage_path: Optional[str] = None,
        storage_mode: Optional[Union[str, PersistentStoreMode]] = None,
        storage_salt: Optional[Union[str, bytes]] = None,
        storage_idlen: Optional[int] = None,
        timezone: Optional[Union[str, tzinfo]] = None,
        **kwargs: Any
    ) -> None:
        """Asset Initialization."""
        # Assign default arguments if specified
        for key, value in kwargs.items():
            if not hasattr(AppriseAsset, key):
                raise AttributeError(
                    f"AppriseAsset init(): An invalid key {key} was specified."
                )

            setattr(self, key, value)

        if plugin_paths:
            # Load any decorated modules if defined
            self.__plugin_paths = plugin_paths
            N_MGR.module_detection(plugin_paths)

        if storage_path:
            # Define our persistent storage path
            self.__storage_path = storage_path

        if storage_mode:
            # Define how our persistent storage behaves
            try:
                self.__storage_mode = (
                    storage_mode if isinstance(storage_mode, NotifyFormat)
                    else PersistentStoreMode(storage_mode.lower())
                )

            except (AttributeError, ValueError, TypeError):
                err = (
                    f"An invalid persistent store mode ({storage_mode}) was "
                    "specified.")
                raise AttributeError(err) from None

        if isinstance(storage_idlen, int):
            # Define the number of characters utilized from our namespace lengh
            if storage_idlen < 0:
                # Unsupported type
                raise ValueError(
                    "AppriseAsset storage_idlen(): Value must "
                    "be an integer and > 0"
                )

            # Store value
            self.__storage_idlen = storage_idlen

        if isinstance(timezone, tzinfo):
            self._tzinfo = timezone

        elif timezone is not None:
            self._tzinfo = zoneinfo(timezone)
            if not self._tzinfo:
                raise AttributeError(
                    "AppriseAsset timezone provided is invalid") from None
        else:
            # Default our timezone to what is detected on the system
            self._tzinfo = datetime.now().astimezone().tzinfo

        if storage_salt is not None:
            # Define the number of characters utilized from our namespace lengh

            if isinstance(storage_salt, bytes):
                self.__storage_salt = storage_salt

            elif isinstance(storage_salt, str):
                try:
                    self.__storage_salt = storage_salt.encode(self.encoding)

                except UnicodeEncodeError:
                    # Bad data; don't pass it along
                    raise ValueError(
                        "AppriseAsset namespace_salt(): "
                        "Value provided could not be encoded"
                    ) from None

            else:  # Unsupported
                raise ValueError(
                    "AppriseAsset namespace_salt(): Value provided must be "
                    "string or bytes object"
                )

    def color(
        self,
        notify_type: NotifyType,
        color_type: Optional[type] = None,
    ) -> Union[str, int, tuple[int, int, int]]:
        """Returns an HTML mapped color based on passed in notify type.

        if color_type is:
           None    then a standard hex string is returned as
                   a string format ('#000000').

           int     then the integer representation is returned
           tuple   then the the red, green, blue is returned in a tuple
        """

        # Attempt to get the type, otherwise return a default grey
        # if we couldn't look up the entry
        color = self.html_notify_map.get(
            notify_type, self.default_html_color)
        if color_type is None:
            # This is the default return type
            return color

        elif color_type is int:
            # Convert the color to integer
            return AppriseAsset.hex_to_int(color)

        # The only other type is tuple
        elif color_type is tuple:
            return AppriseAsset.hex_to_rgb(color)

        # Unsupported type
        raise ValueError(
            "AppriseAsset html_color(): An invalid color_type was specified."
        )

    def ascii(self, notify_type: NotifyType) -> str:
        """Returns an ascii representation based on passed in notify type."""
        # look our response up
        return self.ascii_notify_map.get(
            notify_type, self.default_ascii_chars)

    def image_url(
        self,
        notify_type: NotifyType,
        image_size: Optional[NotifyImageSize] = None,
        logo: bool = False,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Apply our mask to our image URL.

        if logo is set to True, then the logo_url is used instead
        """

        url_mask = self.image_url_logo if logo else self.image_url_mask
        if not url_mask:
            # No image to return
            return None

        if extension is None:
            extension = self.default_extension

        if image_size is None:
            image_size = self.default_image_size

        re_map = {
            "{THEME}": self.theme if self.theme else "",
            "{TYPE}": notify_type.value,
            "{XY}": image_size.value,
            "{EXTENSION}": extension,
        }

        # Iterate over above list and store content accordingly
        re_table = re.compile(
            r"(" + "|".join(re_map.keys()) + r")",
            re.IGNORECASE,
        )

        return re_table.sub(lambda x: re_map[x.group()], url_mask)

    def image_path(
        self,
        notify_type: NotifyType,
        image_size: NotifyImageSize,
        must_exist: bool = True,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Apply our mask to our image file path."""

        if not self.image_path_mask:
            # No image to return
            return None

        if extension is None:
            extension = self.default_extension

        re_map = {
            "{THEME}": self.theme if self.theme else "",
            "{TYPE}": notify_type.value,
            "{XY}": image_size.value,
            "{EXTENSION}": extension,
        }

        # Iterate over above list and store content accordingly
        re_table = re.compile(
            r"(" + "|".join(re_map.keys()) + r")",
            re.IGNORECASE,
        )

        # Acquire our path
        path = re_table.sub(lambda x: re_map[x.group()], self.image_path_mask)
        if must_exist and not isfile(path):
            return None

        # Return what we parsed
        return path

    def image_raw(
        self,
        notify_type: NotifyType,
        image_size: NotifyImageSize,
        extension: Optional[str] = None,
    ) -> Optional[bytes]:
        """Returns the raw image if it can (otherwise the function returns
        None)"""

        path = self.image_path(
            notify_type=notify_type,
            image_size=image_size,
            extension=extension,
        )
        if path:
            try:
                with open(path, "rb") as fd:
                    return fd.read()

            except OSError:
                # We can't access the file
                return None

        return None

    def details(self) -> dict[str, str]:
        """Returns the details associated with the AppriseAsset object."""
        return {
            "app_id": self.app_id,
            "app_desc": self.app_desc,
            "default_extension": self.default_extension,
            "theme": self.theme,
            "image_path_mask": self.image_path_mask,
            "image_url_mask": self.image_url_mask,
            "image_url_logo": self.image_url_logo,
        }

    @staticmethod
    def hex_to_rgb(value: str) -> tuple[int, int, int]:
        """Takes a hex string (such as #00ff00) and returns a tuple in the form
        of (red, green, blue)

        eg: #00ff00 becomes : (0, 65535, 0)
        """
        value = value.lstrip("#")
        lv = len(value)
        return tuple(
            int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3)
        )

    @staticmethod
    def hex_to_int(value: str) -> int:
        """Takes a hex string (such as #00ff00) and returns its integer
        equivalent.

        eg: #00000f becomes : 15
        """
        return int(value.lstrip("#"), 16)

    @property
    def plugin_paths(self) -> list[str]:
        """Return the plugin paths defined."""
        return self.__plugin_paths

    @property
    def storage_path(self) -> Optional[str]:
        """Return the persistent storage path defined."""
        return self.__storage_path

    @property
    def storage_mode(self) -> PersistentStoreMode:
        """Return the persistent storage mode defined."""

        return self.__storage_mode

    @property
    def storage_salt(self) -> bytes:
        """Return the provided namespace salt; this is always of type bytes."""
        return self.__storage_salt

    @property
    def storage_idlen(self) -> int:
        """Return the persistent storage id length."""

        return self.__storage_idlen

    @property
    def tzinfo(self) -> tzinfo:
        """Return the timezone object"""
        return self._tzinfo
